MODULE_INFO = {
    'animation': {'path': 'animation/animation-min.js',
                 'requires': ['dom', 'event'],
                 'type': 'js'},
    'autocomplete': {'optional': ['connection', 'animation'],
                    'path': 'autocomplete/autocomplete-min.js',
                    'requires': ['dom', 'event', 'datasource'],
                    'skinnable': True,
                    'type': 'js'},
    'base': {'after': ['reset', 'fonts', 'grids'],
            'path': 'base/base-min.css',
            'type': 'css'},
    'button': {'optional': ['menu'],
              'path': 'button/button-min.js',
              'requires': ['element'],
              'skinnable': True,
              'type': 'js'},
    'calendar': {'path': 'calendar/calendar-min.js',
                'requires': ['event', 'dom'],
                'skinnable': True,
                'type': 'js'},
    'carousel': {'optional': ['animation'],
                'path': 'carousel/carousel-min.js',
                'requires': ['element'],
                'skinnable': True,
                'type': 'js'},
    'charts': {'path': 'charts/charts-min.js',
              'requires': ['element', 'json', 'datasource'],
              'type': 'js'},
    'colorpicker': {'optional': ['animation'],
                   'path': 'colorpicker/colorpicker-min.js',
                   'requires': ['slider', 'element'],
                   'skinnable': True,
                   'type': 'js'},
    'connection': {'path': 'connection/connection-min.js',
                  'requires': ['event'],
                  'type': 'js'},
    'container': {'optional': ['dragdrop', 'animation', 'connection'],
                 'path': 'container/container-min.js',
                 'requires': ['dom', 'event'],
                 'skinnable': True,
                 'supersedes': ['containercore'],
                 'type': 'js'},
    'containercore': {'path': 'container/container_core-min.js',
                     'pkg': 'container',
                     'requires': ['dom', 'event'],
                     'type': 'js'},
    'cookie': {'path': 'cookie/cookie-min.js',
              'requires': ['yahoo'],
              'type': 'js'},
    'datasource': {'optional': ['connection'],
                  'path': 'datasource/datasource-min.js',
                  'requires': ['event'],
                  'type': 'js'},
    'datatable': {'optional': ['calendar', 'dragdrop', 'paginator'],
                 'path': 'datatable/datatable-min.js',
                 'requires': ['element', 'datasource'],
                 'skinnable': True,
                 'type': 'js'},
    'dom': {'path': 'dom/dom-min.js', 'requires': ['yahoo'], 'type': 'js'},
    'dragdrop': {'path': 'dragdrop/dragdrop-min.js',
                'requires': ['dom', 'event'],
                'type': 'js'},
    'editor': {'optional': ['animation', 'dragdrop'],
              'path': 'editor/editor-min.js',
              'requires': ['menu', 'element', 'button'],
              'skinnable': True,
              'supersedes': ['simpleeditor'],
              'type': 'js'},
    'element': {'path': 'element/element-min.js',
               'requires': ['dom', 'event'],
               'type': 'js'},
    'event': {'path': 'event/event-min.js', 'requires': ['yahoo'], 'type': 'js'},
    'fonts': {'path': 'fonts/fonts-min.css', 'type': 'css'},
    'get': {'path': 'get/get-min.js', 'requires': ['yahoo'], 'type': 'js'},
    'grids': {'optional': ['reset'],
             'path': 'grids/grids-min.css',
             'requires': ['fonts'],
             'type': 'css'},
    'history': {'path': 'history/history-min.js',
               'requires': ['event'],
               'type': 'js'},
    'imagecropper': {'path': 'imagecropper/imagecropper-min.js',
                    'requires': ['dom', 'event', 'dragdrop', 'element', 'resize'],
                    'skinnable': True,
                    'type': 'js'},
    'imageloader': {'path': 'imageloader/imageloader-min.js',
                   'requires': ['event', 'dom'],
                   'type': 'js'},
    'json': {'path': 'json/json-min.js', 'requires': ['yahoo'], 'type': 'js'},
    'layout': {'optional': ['animation', 'dragdrop', 'resize', 'selector'],
              'path': 'layout/layout-min.js',
              'requires': ['dom', 'event', 'element'],
              'skinnable': True,
              'type': 'js'},
    'logger': {'optional': ['dragdrop'],
              'path': 'logger/logger-min.js',
              'requires': ['event', 'dom'],
              'skinnable': True,
              'type': 'js'},
    'menu': {'path': 'menu/menu-min.js',
            'requires': ['containercore'],
            'skinnable': True,
            'type': 'js'},
    'paginator': {'path': 'paginator/paginator-min.js',
                 'requires': ['element'],
                 'skinnable': True,
                 'type': 'js'},
    'profiler': {'path': 'profiler/profiler-min.js',
                'requires': ['yahoo'],
                'type': 'js'},
    'profilerviewer': {'path': 'profilerviewer/profilerviewer-min.js',
                      'requires': ['profiler', 'yuiloader', 'element'],
                      'skinnable': True,
                      'type': 'js'},
    'reset': {'path': 'reset/reset-min.css', 'type': 'css'},
    'reset-fonts': {'path': 'reset-fonts/reset-fonts.css',
                   'rollup': 2,
                   'supersedes': ['reset', 'fonts'],
                   'type': 'css'},
    'reset-fonts-grids': {'path': 'reset-fonts-grids/reset-fonts-grids.css',
                         'rollup': 4,
                         'supersedes': ['reset', 'fonts', 'grids', 'reset-fonts'],
                         'type': 'css'},
    'resize': {'optional': ['animation'],
              'path': 'resize/resize-min.js',
              'requires': ['dom', 'event', 'dragdrop', 'element'],
              'skinnable': True,
              'type': 'js'},
    'selector': {'path': 'selector/selector-min.js',
                'requires': ['yahoo', 'dom'],
                'type': 'js'},
    'simpleeditor': {'optional': ['containercore',
                                'menu',
                                'button',
                                'animation',
                                'dragdrop'],
                    'path': 'editor/simpleeditor-min.js',
                    'pkg': 'editor',
                    'requires': ['element'],
                    'skinnable': True,
                    'type': 'js'},
    'slider': {'optional': ['animation'],
              'path': 'slider/slider-min.js',
              'requires': ['dragdrop'],
              'skinnable': True,
              'type': 'js'},
    'stylesheet': {'path': 'stylesheet/stylesheet-min.js',
                  'requires': ['yahoo'],
                  'type': 'js'},
    'tabview': {'optional': ['connection'],
               'path': 'tabview/tabview-min.js',
               'requires': ['element'],
               'skinnable': True,
               'type': 'js'},
    'treeview': {'optional': ['json'],
                'path': 'treeview/treeview-min.js',
                'requires': ['event', 'dom'],
                'skinnable': True,
                'type': 'js'},
    'uploader': {'path': 'uploader/uploader.js',
                'requires': ['element'],
                'type': 'js'},
    'utilities': {'path': 'utilities/utilities.js',
                 'rollup': 8,
                 'supersedes': ['yahoo',
                               'event',
                               'dragdrop',
                               'animation',
                               'dom',
                               'connection',
                               'element',
                               'yahoo-dom-event',
                               'get',
                               'yuiloader',
                               'yuiloader-dom-event'],
                 'type': 'js'},
    'yahoo': {'path': 'yahoo/yahoo-min.js', 'type': 'js'},
    'yahoo-dom-event': {'path': 'yahoo-dom-event/yahoo-dom-event.js',
                       'rollup': 3,
                       'supersedes': ['yahoo', 'event', 'dom'],
                       'type': 'js'},
    'yuiloader': {'path': 'yuiloader/yuiloader-min.js',
                 'supersedes': ['yahoo', 'get'],
                 'type': 'js'},
    'yuiloader-dom-event': {'path': 'yuiloader-dom-event/yuiloader-dom-event.js',
                           'rollup': 5,
                           'supersedes': ['yahoo',
                                         'dom',
                                         'event',
                                         'get',
                                         'yuiloader',
                                         'yahoo-dom-event'],
                           'type': 'js'},
    'yuitest': {'path': 'yuitest/yuitest-min.js',
               'requires': ['logger'],
               'skinnable': True,
               'type': 'js'}}

SKIN = {
    'after': ['reset', 'fonts', 'grids', 'base'],
    'base': 'assets/skins/',
    'defaultSkin': 'sam',
    'path': 'skin.css',
    'rollup': 3}

MODULE_INFO[SKIN['defaultSkin']] = {
    'type': 'css',
    'path': SKIN['base'] + SKIN['defaultSkin'] + '/' + SKIN['path'],
    'after': SKIN['after'] }



import re
from shlex import shlex

from django.conf import settings

from yui_loader.module_info_2_7_0 import MODULE_INFO, SKIN
from yui_loader.components import Components


DEFAULT_BASE = 'http://yui.yahooapis.com/2.7.0/build/'
YUI_BASE = getattr(settings, 'YUI_INCLUDE_BASE', DEFAULT_BASE)

VERSIONS = {'raw': '', '': '', 'min': '-min', 'debug': '-debug'}

DEFAULT_JS_TAG = '<script type="text/javascript" src="%s"></script>'
DEFAULT_CSS_TAG = '<link rel="stylesheet" type="text/css" href="%s" />'
TAGS = {'js': getattr(settings, 'YUI_INCLUDE_JS_TAG', DEFAULT_JS_TAG),
        'css': getattr(settings, 'YUI_INCLUDE_CSS_TAG', DEFAULT_CSS_TAG)}

class YUILoader:

    def __init__(self, *components, **kwargs):
        self._module_info = Components(MODULE_INFO)
        self._components = set()
        self._rolled_up_components = {}
        self._rollup_counters = {}
        self.set_version(kwargs.pop('version', 'min'))
        for module in kwargs.pop('add_modules', ()):
            self.add_module(module)
        for component in components:
            self.add_component(component)

    def set_version(self, version):
        self._version = VERSIONS[version]

    def add_component(self, new_component_name):
        if not self._has_component(new_component_name):
            self._add_requirements(new_component_name)
            self._count_in_rollups(new_component_name)
            rollup_name = self._get_satisfied_rollup(new_component_name)
            if rollup_name:
                self.add_component(rollup_name)
            else:
                self._components.add(new_component_name)
                self._roll_up_superseded(new_component_name)

    def add_module(self, module_def):
        module_data = {}
        lexer = shlex(module_def, posix=True)

        def expect(*patterns):
            token = lexer.get_token()
            if token not in patterns:
                raise ValueError, '%s expected instead of %s' % \
                      (' or '.join(repr(s) for s in patterns),
                       token and repr(token) or 'end of data')
            return token

        str_attrs = 'name', 'type', 'path', 'fullpath', 'varName'
        list_attrs = 'requires', 'optional', 'after'
        state = 'ATTR'
        expect('{')
        while state != 'STOP':
            if state == 'ATTR':
                token = expect(*str_attrs+list_attrs)
                expect(':')
                if token in str_attrs:
                    module_data[token] = lexer.get_token()
                    if module_data[token] is None:
                        raise ValueError, \
                              'string expected instead of end of data'
                    state = 'DELIM'
                elif token in list_attrs:
                    expect('[')
                    lst = module_data[token] = []
                    state = 'LIST'
            elif state == 'LIST':
                lst.append(lexer.get_token())
                if re.search(r'\W', lst[-1]):
                    raise ValueError, 'invalid component name %r' % token
                if expect(',', ']') == ']':
                    state = 'DELIM'
            elif state == 'DELIM':
                if expect(',', '}') == '}':
                    expect(None)
                    state = 'STOP'
                else:
                    state = 'ATTR'

        if 'type' not in module_data:
            raise ValueError, 'type missing in %r' % module_def
        self._module_info.add(module_data['name'], module_data)
        return module_data

    def render(self):
        return '\n'.join(self._render_component(component)
                         for component in self._sort_components())

    def _urls_by_type(self, component_type):
        sorted_components = (self._module_info[component_name]
                             for component_name in self._sort_components())
        return [self._component_path(component)
                for component in sorted_components
                if component.type == component_type]

    def js_urls(self):
        return self._urls_by_type('js')

    def css_urls(self):
        return self._urls_by_type('css')

    def _has_component(self, component_name):
        return component_name in self._components \
               or component_name in self._rolled_up_components

    def _get_satisfied_rollup(self, component_name):
        if self._version == '-min':
            for rollup_name in self._module_info.get_rollups(component_name):
                rollup_status = self._rollup_counters.get(rollup_name, set())
                if len(rollup_status) >= self._module_info[rollup_name].rollup:
                    return rollup_name

    def _count_in_rollups(self, component_name):
        for rollup_name in self._module_info.get_rollups(component_name):
            rolled_up = self._rollup_counters.setdefault(rollup_name, set())
            rolled_up.add(component_name)
        for superseded in self._module_info[component_name].supersedes:
            self._count_in_rollups(superseded)

    def _roll_up_superseded(self, component_name):
        for superseded in self._module_info[component_name].supersedes:
            self._rolled_up_components[superseded] = component_name
            if superseded in self._components:
                self._components.remove(superseded)

    def _add_requirements(self, component_name):
        component = self._module_info[component_name]
        for requirement in component.requires:
            self.add_component(requirement)
        if component.skinnable:
            self.add_component(SKIN['defaultSkin'])

    def _component_path(self, component):
        path = component.fullpath or YUI_BASE + component.path
        if component.type == 'js':
            if self._version != '-min' and path.endswith('-min.js'):
                path = path[:-7] + self._version + '.js'
        elif component.type == 'css':
            if self._version == '' and path.endswith('-min.css'):
                path = path[:-8] + '.css'
        return path

    def _render_component(self, component_name):
        component = self._module_info[component_name]
        return TAGS[component.type] % self._component_path(component)

    def _sort_components(self, component_names=None):
        if component_names is None:
            comps = self._components.copy()
        else:
            comps = component_names
        while comps:
            component_name = comps.pop()
            component = self._module_info[component_name]
            direct_deps = component.requires + component.after
            indirect_deps = [
                self._rolled_up_components[r] for r in direct_deps
                if r in self._rolled_up_components]
            all_deps = set(direct_deps) \
                       .union(set(indirect_deps)) \
                       .union(set(component.optional))
            deps_left = comps.intersection(all_deps)
            for r in self._sort_components(deps_left):
                yield r
                comps.remove(r)
            yield component_name



__doc__ = """

YUI_include -- YUI Loader as Django middleware

(c) Antti Kaihola 2008,2009  http://djangopeople.net/akaihola/
                             akaihol+django@ambitone.com

This server-side middleware implements some of the functionality in the Yahoo
User Interface Loader component.  YUI JavaScript and CSS modules requirements
can be declared anywhere in the base, inherited or included templates, and the
resulting, optimized <script> and <link rel=stylesheet> tags are inserted at
the specified position of the resulting page.

Requirements may be specified in multiple locations.  This is useful when zero
or more components are included in the HTML head section, and inherited and/or
included templates require possibly overlapping sets of YUI components in the
body across inherited and included templates.  All tags are collected in the
head section, and duplicate tags are automatically eliminated.

The middleware understands component dependencies and ensures that resources
are loaded in the right order.  It knows about built-in rollup files that ship
with YUI.  By automatically using rolled-up files, the number of HTTP requests
is reduced.

The default syntax looks like HTML comments.  Markup for the insertion point is
replaced with <script> and <link> tags:
<!-- YUI_init -->

Component requirements are indicated, possibly in multiple locations, with the
``YUI_include`` markup.  It is removed from the resulting page by the
middleware. Example:
<!-- YUI_include fonts grids event dragdrop -->

Non-minified and compressed versions are requested, respectively, by:
<!-- YUI_version raw -->
<!-- YUI_version debug -->

Example:

<html><head>
<!-- YUI_init -->
<!-- YUI_include dom event -->
</head><body>
<!-- YUI_include element selector reset fonts base -->
</body></html>

Renders:

<html><head>
<link rel="stylesheet" type="text/css" href="http://yui.yahooapis.com/2.7.0/build/reset-fonts/reset-fonts.css" />
<link rel="stylesheet" type="text/css" href="http://yui.yahooapis.com/2.7.0/build/base/base-min.css" />
<script type="text/javascript" src="http://yui.yahooapis.com/2.7.0/build/yahoo-dom-event/yahoo-dom-event.js"></script>
<script type="text/javascript" src="http://yui.yahooapis.com/2.7.0/build/element/element-beta-min.js"></script>
<script type="text/javascript" src="http://yui.yahooapis.com/2.7.0/build/selector/selector-beta-min.js"></script>
</head><body>
</body></html>

The markup format can be customized with global Django settings.  Example:
YUI_INCLUDE_PREFIX_RE = r'{!'
YUI_INCLUDE_SUFFIX_RE = r'!}'
would change markup to e.g. ``{! init !}`` and ``{! include dom event !}``.

The base URL is customized with the ``YUI_INCLUDE_BASE`` setting, e.g.:
YUI_INCLUDE_BASE = 'http://localhost:8000/yui/build/'

To remove the XHTML trailing slash from the <link> tag, use:
YUI_INCLUDE_CSS_TAG = '<link rel="stylesheet" type="text/css" href="%s">'
"""

import re
from django.conf import settings
from yui_loader.loader import YUILoader

PREFIX_RE = getattr(settings, 'YUI_INCLUDE_PREFIX_RE', '<!-- *YUI_')
SUFFIX_RE = getattr(settings, 'YUI_INCLUDE_SUFFIX_RE', ' *-->')

YUI_RE = re.compile(
    r'%s(include|version) +(.*?)%s' % (PREFIX_RE, SUFFIX_RE))
YUI_ADDMODULE_RE = re.compile(
    r'(?s)%saddModule\s*(\{\s*.*?\s*})\s*%s' % (PREFIX_RE, SUFFIX_RE))
YUI_INIT_RE = re.compile(
    '%sinit%s' % (PREFIX_RE, SUFFIX_RE))

ACCEPTED_CONTENT_TYPES = ('text/html',
                          'text/xml',
                          'application/xhtml+xml',
                          'application/xml')

class YUIIncludeMiddleware(object):
    def process_response(self, request, response):
        if getattr(response, 'disable_yui_loader_middleware', False):
            # TODO: support Django 1.5's  StreamingHttpResponse too
            return response

        content_type = response['Content-Type'].split(';')[0]
        if content_type not in ACCEPTED_CONTENT_TYPES:
            return response

        components = set()
        loader = YUILoader()

        def add_module(match):
            loader.add_module(match.group(1))
            return ''
        content = YUI_ADDMODULE_RE.sub(add_module, response.content)

        def collect(match):
            cmd, data = match.groups()
            if cmd == 'include':
                components.update(data.split())
            elif cmd == 'version':
                loader.set_version(data)
            else:
                return '<!-- UNKNOWN COMMAND YUI_%s -->' % cmd
            return ''
        content = YUI_RE.sub(collect, content)

        for component in components:
            loader.add_component(component)

        tags = loader.render()
        if tags:
            content, count = YUI_INIT_RE.subn(tags, content, 1)
            if count != 1:
                content += ('<p>%d YUI init tags found,'
                            'at least one expected</p>' % count)
            response.content = YUI_INIT_RE.sub(
                '<!-- WARNING: MULTIPLE YUI init STATEMENTS -->', content)

        return response



class ComponentAdapter:

    def __init__(self, components, name, data):
        self.components = components
        self.name = name
        self.data = data

    @property
    def optional(self):
        return self.data.get('optional', [])

    @property
    def supersedes(self):
        return self.data.get('supersedes', [])

    @property
    def requires(self):
        r = self.data.get('requires', [])

        if self.type != 'js' \
        or self.name == 'yahoo' \
        or 'yahoo' in self.supersedes \
        or 'yahoo' in r:
            return r
        else:
            return ['yahoo'] + r

    @property
    def after(self):
        if self.type == 'css':
            return self.data.get('after', [])
        elif self.type == 'js':
            return self.data.get('after', []) + \
                   self.components.get_css_components()

    @property
    def rollup(self):
        return self.data.get('rollup', 1)

    @property
    def skinnable(self):
        return self.data.get('skinnable', False)

    @property
    def fullpath(self):
        return self.data.get('fullpath', None)

    def __getattr__(self, attname):
        return self.data[attname]


class Components(dict):
    def __init__(self, components_dict):
        super(Components, self).__init__()
        self.rollup_mapping = {}
        for component_name, data in components_dict.items():
            self.add(component_name, data)

    def __getitem__(self, component_name):
        data = super(Components, self).__getitem__(component_name)
        return ComponentAdapter(self, component_name, data)

    def get_rollups(self, component_name):
        return self.rollup_mapping[component_name]

    def get_all_rollups(self, component_name):
        rollups = self.get_rollups(component_name)
        for superseded in self[component_name].supersedes:
            rollups.update(self.get_all_rollups(superseded))
        return rollups

    def get_css_components(self):
        return [name for name, data in self.items()
                if data['type'] == 'css']

    def add(self, component_name, data):
        self[component_name] = data
        self.rollup_mapping.setdefault(component_name, set())
        if 'rollup' not in data:
            return
        for rolled_up in data['supersedes']:
            self.rollup_mapping.setdefault(rolled_up, set()).add(
                component_name)



__doc__ = """

YUI_include -- YUI Loader as Django middleware

<<<<<<< HEAD
(c) 2008-2009 Antti Kaihola and individual contributors
              akaihol+django@ambitone.com
=======
(c) Antti Kaihola 2008,2009,2013
>>>>>>> 9a7c7b3... Added detection for `response.disable_yui_loader_middleware = True`

License: BSD, see the file LICENSE for details

See README.rst for documentation
"""

import ez_setup
ez_setup.use_setuptools()
from setuptools import setup

setup(
    name='django-yui-loader',
    author='Antti Kaihola',
    author_email='akaihol+django@ambitone.com',
    maintainer='Antti Kaihola',
    maintainer_email='akaihol+django@ambitone.com',
    version='0.3.2',
    url='http://github.com/akaihola/django-yui-loader',
    packages=['yui_loader'],
    description=('Server-side middleware which implements some of the '
                 'functionality in the Yahoo User Interface Loader '
                 'component.'),
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Utilities'
    ]
)
